// file: kernel.cpp
// needs: dib.*, image.*, lib1.h, msgbox.h, krnldef.h, mathkrnl.*, kernel.h
// Kernel functions for fractal image encoding process
// DIP kernel
// IFS module
// Author: Giovambattista Pulcini
// Contributes by: Riccardo Rossi, Valerio Verrando
// Written: 1994.12.10
// Revised:
//      5.10 - Oct 1995 - 32bit Flat memory model & NT port
//      5.03 - Apr 1995 - GetTimeString
//      5.02 - Apr 1995 - WriteCoeff, ReadCoeff of polynomials
//      5.01 - Mar 1995 - Header, WriteHeader
// older versions:
//      5.0  * Introducing RMS, DeltaRMS and SNR in decoding
//      4.9  * Updating Decode4 for new file format xx03
//      4.8  * Operating orientation scheme
//      4.7  * Support for orientation scheme
//      4.6  * Dialog with most settings
//      4.5  * Local Sub Search
//      4.4  * Added a Status window multiple file encoding, decoding from file
//      4.3  * fixed bug on alpha,beta quantization, reduced bits for encoding domain position
//      4.2  * Weights implementation
//      4.1  * alpha, beta quantization.
//      4.0  * fixed bug on multiple resolution decoding: no UAE in destroing CODE tree
//             implemented single-step saving
//      3.9  * fixed bug on multiple resolution decoding.
//      3.8  * implemented multiple resolution decoding: buggy
//      3.7  * fixed bug on optimal alpha,beta storing in code-file.

#define KERNELVER "5.10"

/*

To do list:
  add a CS_NULL classification scheme
  Review MakeDomainPool for max speed

*/

#include "kernel.h"
#include "mem.h"

#include "report.cpp"

// Globals
char far * ClassSchemeNames[15] = {"Histogram", "Gradient", "VSS", "Tween Peaks"};
char far * BlockTypeNames[15] = {"Shade", "Midrange", "Edge", "Multiple Edge" };
char far * BlockTypeShorts[4] = {"SH", "MR", "E", "ME" };
char far * KrnlMetricsNames[10] = {"MS","RMS","SNR","SNRMS","SNRRMS","SNRMSdB","SNRRMSdB","MABS","MAX","MMAX" };
char far * KrnlApproxNames[10] =  {"deg0","deg1","deg2","bilin","biquad"};

//enum TBlockType { SHADE, MIDRANGE, EDGE, MULTIEDGE }; //int
//enum TClassScheme{ CS_HISTOGRAM, CS_GRADIENT, CS_VSS, CS_TWEENPEAKS }; //int
//enum TMetrics { MS, RMS, SNR, SNRMS, SNRRMS, SNRMSdB, SNRRMSdB, ABSM, MAX, MAXM }; // int

//////////////////////////////////////////////////////////////////////
// Kernel

#define GRAYLEVELS 256
#define MAXGRAY    255

#define FORALLIsometries(x) \
	for( x=iso_IDENTITY;  x<=iso_LAST; \
	     x = TIsometries((x) + TIsometries(1)) )

char * Kernel::GetTimeString()
{
/*
  time_t t;
  struct time d_time;
  struct date d_date;
  struct tm *local;

  getdate(&d_date);
  gettime(&d_time);

  t = dostounix(&d_date, &d_time);
  local = localtime(&t);
*/

  time_t t;
  t = time( NULL );
  struct tm * local;
  local = localtime(&t);
  return asctime(local);
}

DWORD Kernel::MakeHistogram(const RTImage img, DWORD * Histogram)
{
    ShowBegin();
    ShowMessage("Making Histogram");
    ShowRatio(0);
    DWORD Time = GetTimeTicks();
    DWORD max = 0;
    for (int k=0; k<256; k++) Histogram[k]=0;
    for (WORD i=0; i<img.wHeight; i++)
    {
      if (i%10) ShowRatio(float(i)/float(img.wHeight-1));
      for (WORD j=0; j<img.wWidth; j++)
	{
	  DWORD x = Histogram[img.Pixel(i,j)]++;
	  max = Max(max, x);
	}
    }
    Timing = GetTimeTicks()-Time;
    ShowEnd();
    return max;
}

void Kernel::ImageBlockClassify(const RTImage img, DWORD Blocks[4])
{
    ShowBegin();
    ShowMessage("Classifying Blocks...");
    ShowRatio(0);
    DWORD Time = GetTimeTicks();

    WORD imax = img.wHeight/B + (img.wHeight%B ? 1:0);
    WORD jmax = img.wWidth/B + (img.wWidth%B ? 1:0);
    for (int k=0; k<4; k++) Blocks[k]=0;
    for (WORD ib=0; ib<imax; ib++)
      for (WORD jb=0; jb<jmax; jb++)
	{
	  ShowRatio(double(ib*jmax+jb)/(imax*jmax));
	  // Here we selected the block S[ib,jb,B] of size BxB
	  Blocks[BlockClassify(img, ib,jb)]++;
	}
    Timing = GetTimeTicks()-Time;
    ShowEnd();
}


DWORD Kernel::GetDPoolSize(const RTImage img)
// returns byte size of D-Pool for img
{
    WORD nd,md;
    GetDomainBlocksNumber(img, nd, md);
    return(((DWORD)nd)*md*B*B);
}

DWORD Kernel::GetDPoolInfoSize(const RTImage img)
// returns byte size of D-Pool info records for img
{
    WORD nd,md;
    GetDomainBlocksNumber(img, nd, md);
    return(((DWORD)nd)*md*sizeof(TDpool));
}

DWORD Kernel::GetRPoolInfoSize(const RTImage img)
// returns byte size of R-Pool info records for img
{
    WORD nr,mr;
    GetRangeBlocksNumber(img, nr, mr);
    return(((DWORD)nr)*mr*sizeof(TRpool));
}

void Kernel::MakeDomainPool(const RTImage img, BYTE    * DPool, TDpool    * DPoolInfo, DWORD &DPoolSize, DWORD Blocks[4])
// Caller must allocate GetDPoolSize bytes for DPool
// and DPoolInfoSize bytes of DPoolInfo
{
    if(!CompatibleImage(img)) MsgBox("Error:","ill conditioned image.");

    #ifdef DEBUG_MDP_REP
    FILE * rpt = fopen("Dpool.rpt", "w");
    #endif
    WORD nd,md;
    GetDomainBlocksNumber(img, nd, md);
    if (Verbose)
      MsgBox("Domain Pool",
	     "Domain cell size: %2d\n"
	     "Range cell size : %2d\n"
	     "Cells(n,m): %d,%d\n"
	     "Block size: %4ld KB\n"
	     "Info size : %4ld KB",
	     D,B,nd,md,((DWORD)nd)*md*B*B/1024, ((DWORD)nd)*md*sizeof(TDpool)/1024);

    ShowBegin();
    ShowMessage("Making Domain pool...");
    ShowRatio(0);
    DWORD Time = GetTimeTicks();
    BYTE * RBlock = new BYTE[B*B];

    if (Blocks) { for (int k=0; k<4; k++) Blocks[k]=0; }

    DWORD k=0; // number of effective D-Pool blocks : is returned as DPoolSize
    for (WORD id=0; id<nd; id++)
      {
      ShowRatio(double(id)/nd);
      for (WORD jd=0; jd<md; jd++)
	{
	  Contract(img, id, jd, &DPool[k*(B*B)], (BYTE    *) &DPoolInfo[k]);
	  #ifdef DEBUG_MDP_REP
	  fprintf(rpt, "%d,%d: d:%lf d2:%lf\n",id*did,jd*djd,DPoolInfo[k].WSum, DPoolInfo[k].WSum2); fflush(rpt);
	  #endif
	  if (Blocks) Blocks[DPoolInfo[k].BlockType]++;
	  if (ENC_STRIPSHADES)
	      { //if (DPoolInfo[k].BlockType != SHADE) k++;
		// We try to approximate Block with a polynomial of the same degree of
		// luminance shift term. If error is under a threshold we discard block
		// from domain
		//double E = Approximate(&DPool[k*(B*B)], RBlock, B, p);
		//if (E>=ErrorThreshold) k++; // do not discard this block
		// Instead of stripping we can mark these domains.
		k++;
	      }
	  else k++;
	}
      }
    if (k==0) k++; // at least 1 domain block in pool, signal this situation!!
    DPoolSize = k;
    delete RBlock;

    Timing = GetTimeTicks()-Time;
    ShowEnd();
    #ifdef DEBUG_MDP_REP
    fclose(rpt);
    #endif
}

void Kernel::MakeRangePool(const RTImage img, TRpool * RPoolInfo, DWORD Blocks[4])
// Caller must allocate GetRPoolInfoSize bytes for RPool
{

    if(!CompatibleImage(img)) MsgBox("Error:","ill conditioned image.");

    #ifdef DEBUG_MRP_REP
    FILE * rpt = fopen("Rpool.rpt","w");
    #endif

    WORD nr,mr;
    GetRangeBlocksNumber(img, nr, mr);
    if (Verbose)
      MsgBox("Range Pool",
	   "Range cell size : %2d\n"
	   "Cells(n,m): %d,%d\n"
	   "Info size : %4ld KB",
	   B,nr,mr,((DWORD)nr)*mr*sizeof(TDpool)/1024);

    ShowBegin();
    ShowMessage("Making Range pool...");
    ShowRatio(0);
    DWORD Time = GetTimeTicks();

    WORD B2=B/2;
    THistogram Histogram;
    if (Blocks) { for (int k=0; k<4; k++) Blocks[k]=0; }
    for (WORD ir=0; ir<nr; ir++)
      {
      ShowRatio(double(ir)/nr);
      for (WORD jr=0; jr<mr; jr++)
	{
	  RPoolInfo[ir*mr+jr].MeanGray = BlockHistogram(img, ir,jr, Histogram);
	  RPoolInfo[ir*mr+jr].BlockType = BlockClassify(img, ir, jr);
	  double WSum, WSum2, SSum[4]; BlockSum(img, ir, jr, WSum, WSum2, SSum);
	  RPoolInfo[ir*mr+jr].WSum = WSum;
	  RPoolInfo[ir*mr+jr].WSum2 = WSum2;
	  RPoolInfo[ir*mr+jr].SubMean[0] = 0.5+SSum[0]/(B2*B2);
	  RPoolInfo[ir*mr+jr].SubMean[1] = 0.5+SSum[1]/(B2*B2);
	  RPoolInfo[ir*mr+jr].SubMean[2] = 0.5+SSum[2]/(B2*B2);
	  RPoolInfo[ir*mr+jr].SubMean[3] = 0.5+SSum[3]/(B2*B2);

	  // ordering luminance sub-blocks 0123 index in a[] induce crescent order
	  // a[0] most bright sub-block, ..., a[3] most dark sub-block (or its derivative)
	  int i,k,a[4]={0,1,2,3};
	  for(k=0; k<3; k++)
	    for(i=k+1; i<4; i++)
	      if (SSum[a[k]]<SSum[a[i]]) Swap(a[k],a[i]);
	  RPoolInfo[ir*mr+jr].Orientation = Permutation(a,4);  // gets lexicographic order of perm.
	  // NOTE: we can calculate lexicographic order while ordering, simply by
	  // counting swaps we make.

	  #ifdef DEBUG_MRP_REP
	  fprintf(rpt, "%d,%d r:%lf r2:%lf\n",ir,jr,WSum,WSum2); fflush(rpt);
	  #endif
	  if (Blocks) Blocks[RPoolInfo[ir*mr+jr].BlockType]++;
	}
      }

    Timing = GetTimeTicks()-Time;
    
    ShowEnd();
    
    #ifdef DEBUG_MRP_REP
    fclose(rpt);
    #endif
}

TBlockType Kernel::BlockClassify(const RTImage img, WORD ib0, WORD jb0)
// Do not use this function, it will be suppressed and will not be supported in
// future release.
// Extract a block from image by GetImageBlock function and then classify it by
// calling Classify, described in the following.
// * Works only on BxB blocks. ib0,jb0 are in range 0..(H/B); 0..(W/B)
// * (0,0) block is @ the lower left corner of the image.
{
  BYTE * Block = new BYTE[B*B];
  img.GetBlock(ib0*B, jb0*B, B, Block);
  TBlockType temp = Classify(Block, B);
  delete Block;
  return temp;
}

TBlockType Kernel::Classify(BYTE    * Bits, WORD N)
// This function classifies the block NxN pointed by Bits.
// Change this to change classification algorithm.
// Remember to keep this as efficient as you can, it's used inside some nested loops.
{
  switch(ClassScheme)
  {
  case CS_HISTOGRAM:
    {
    //WORD Histogram[256];
    //memset(&Histogram, 0x0, sizeof(Histogram));  // very fast clear
    WORD I;
    BYTE Hmin, Hmax;
    Hmin=Hmax=Bits[0];
    for (WORD i=0; I=N*i,i<N; i++)
      for (WORD j=0; j<N; j++)
	{
	 //Histogram[Bits[I+j]]++;
	 BYTE x = Bits[I+j];
	 if (x<Hmin) Hmin=x;
	 if (x>Hmax) Hmax=x;
       }
    BYTE dr = Hmax-Hmin;
    if (dr==0) return SHADE;
    return TBlockType(dr/64);
    /* We assume the following coarse classification:
       dr : 0..63 SHADE
       dr : 64..127 MIDRANGE
       dr : 128..195 SIMPLE EDGE
       dr : 196..255 MULTI EDGE
    */
    }
  case CS_TWEENPEAKS:
    {
    WORD max1,max2;
    THistogram Histogram; memset(Histogram, 0x0, sizeof(Histogram));
    //for (WORD k=0; k<256; k++) Histogram[k]=0;
    WORD I,k;
    for (WORD i=0; I=N*i,i<N; i++)
      for (WORD j=0; j<N; j++)
	Histogram[Bits[I+j]]++;
    max1 = max2 = 0; // indexes
    for (k=0; k<256; k++)
      {
	  BYTE x = Histogram[k];
	  if (x>Histogram[max1]) max2=max1, max1=k;
	  else if (x>Histogram[max2]) max2=k;
      }
    double d = abs(int(max1)-int(max2))/255; // is the distance (in gray-levels) between two peaks
    // The following 16, 64 are just casual; try better values by visual inspection
    /*if (d<16) return SHADE;
    if (d<64) return MIDRANGE;
    return EDGE;*/
    if (d<ShadeThreshold) return SHADE;
    if (d<MidrangeThreshold) return MIDRANGE;
    return EDGE;
    }
  case CS_GRADIENT:
  case CS_VSS:
    {
    double e,p;
    VSSprojection1(Bits, N, e, p);
    //return TBlockType( int(e*1000)/100 ); // e1000/100
    if (e < ShadeThreshold) return SHADE;
    if (e < MidrangeThreshold) return MIDRANGE;
    return EDGE;
    }
  default:
    break;
  }
  return SHADE; //dummy
}

void Kernel::Contract(const RTImage img, WORD id0, WORD jd0, BYTE    * DPool, BYTE    * DPoolInfo)
{
    WORD IB, B2 = B/2, IB2,
	 id = id0*did,
	 jd = jd0*djd;
    double WSum = 0, WSum2 = 0, Sum = 0;
    double SSum[4]={0.0,0.0,0.0,0.0};
    double SSum2[4]={0,0,0,0};
    for (WORD i=0; IB=B*i, IB2=i/B2*2, i<B; i++)
      for (WORD j=0; j<B; j++)
	{
	  WORD I = id+(i<<1),  // id+2i
	       J = jd+(j<<1);  // jd+2j
	  double x =((double)img.Pixel(I,J)+(double)img.Pixel(I+1,J)+(double)img.Pixel(I,J+1)+(double)img.Pixel(I+1,J+1))*0.25;
	  BYTE  bx = BYTE(Bound(0.5+x,0.0,255.0));  // Bound could be unnecessary
	  DPool[IB+j] = bx;
	  double wx = Weight[IB+j]*bx;
	  Sum  += x;
	  WSum += wx;
	  WSum2+= wx*bx;
	  SSum[IB2+j/B2] += x;
	  SSum[IB2+j/B2] += x*x;
	}
    TDpool    * DP = (TDpool    *) DPoolInfo;
    if (DP)
      {
	DP->MeanGray = 0.5+Sum/(B*B);
	DP->BlockType= Classify(DPool, B);
	DP->id = id0;
	DP->jd = jd0;
	DP->WSum = WSum;
	DP->WSum2= WSum2;
	DP->SubMean[0] = SSum[0];
	DP->SubMean[1] = SSum[1];
	DP->SubMean[2] = SSum[2];
	DP->SubMean[3] = SSum[3];

	// ordering luminance sub-blocks 0123 index in a[] induce crescent order
	// a[0] most bright sub-block, ..., a[3] most dark sub-block (or its derivative)
	int i,k,a[4]={0,1,2,3};
	for(k=0; k<3; k++)
	  for(i=k+1; i<4; i++)
	    if (SSum[a[k]]<SSum[a[i]]) Swap(a[k],a[i]);
	DP->Orientation = Permutation(a,4);  // gets lexicographic order of perm.
	// NOTE: we can calculate lexicographic order while ordering, simply by
	// counting swaps we make.
      }
}


double Kernel::Distance(BYTE    * A, BYTE    * B, WORD N)
// Note: for SNR A is Signal, B its approximation or noisy representation
{
  WORD I,i,j;
  double d=0,f=0,dr=0;
  BYTE Hmin, Hmax, x;
  switch(Metric)
  {
  case SNR:
      Hmin=Hmax=A[0];
      for(i=0; I=i*N,i<N; i++)
	for(j=0; j<N; j++)
	  {  d += sqr(double(x=A[I+j])-B[I+j]);
	     if (x<Hmin) Hmin=x;
	     if (x>Hmax) Hmax=x;
	  }
      dr = Hmax-Hmin;
      break;
  case MS:
  case RMS:
      for(i=0; I=i*N,i<N; i++)
	for(j=0; j<N; j++)
	  d += sqr(double(A[I+j])-B[I+j]);
      break;
  case SNRMS:
  case SNRRMS:
  case SNRRMSdB:
  case SNRMSdB:
      for(i=0; I=i*N,i<N; i++)
	for(j=0; j<N; j++)
	  { d += sqr(double(A[I+j])-B[I+j]);
	    f += sqr(double(B[I+j]));
	  }
      break;
  case ABSM:
      for(i=0; I=i*N,i<N; i++)
	for(j=0; j<N; j++)
	  d += abs(double(A[I+j])-B[I+j]);
      break;
  case MAXM:
  case MAX:
      Hmin=Hmax=abs(A[0]-B[0]);
      for(i=0; I=i*N,i<N; i++)
	for(j=0; j<N; j++)
	  {
	     x = abs(A[I+j]-B[I+j]);
	     if (x<Hmin) Hmin=x;
	     if (x>Hmax) Hmax=x;
	  }
      break;

  }
  switch(Metric)
  {
    case MS:     // Mean Square
      return d/(N*N);
    case RMS:    // Root Mean Square
      return sqrt(d/(N*N));
    case SNRMS:  // Signal/Noise Ratio Mean Square
      if (d!=0) return f/d; else return -1; // -1 SNR=Infinity
    case SNRRMS: // Signal/Noise Ratio Root Mean Square
      if (d!=0) return sqrt(f/d); else return -1; // -1 SNR=Infinity
    case SNRMSdB:  // SNR MS in dB
      if (d!=0) return 10*log10(f/d); else return -1; // Warning -1 may be good value
    case SNRRMSdB: // SNR RMS in dB
      if (d!=0) return 5*log10(f/d); else return -1;
    case SNR:
      if (d!=0) return 10*log10(sqr(dr)/(d*(N*N))); else return -1;
    case ABSM:
      return d/(N*N);
    case MAX:
      return Hmax;
    case MAXM:
      return double(Hmax)/(N*N);
    default: ;
  }
  return -1; //dummy
}

double Kernel::Approximate(PTBLOCK A, PTBLOCK B, WORD N, double * p)
// A is block which is to be approximate by B of size NxN.
// Approximation type is setted by global ApproximationType
// Returns Global error according to Metric
{
  double * b = new double[N*N];
  double m = 0;
  WORD k;
  switch (ApproximationType)
    {
    case APPT_DEG0:
	for(k=0; k<N*N; k++) m += (b[k]=A[k]);
	m/=(N*N);
	for(k=0; k<N*N; k++) //B[k]=quant(m+0.5,0,255,6); // Round
	B[k] = int((m/255.0)*64)*255/64; // quantization to 6 bits over 8
	p[0]=B[0]; //quant(m+0.5,0,255,6);
	break;
    case APPT_DEG1:
	break;
    case APPT_DEG2:
	OPENREPORT("Blk2.rpt");
	//fprintf(rpt, "Block:\n");
	for(k=0; k<N*N; k++) b[k]=A[k];
	// WriteMatrix(rpt,N,N,b);
	double ab[6];
	for(k=0; k<6; k++)
	  ab[k] = DotVector(N*N,&Basis[k*N*N],b,Weight);
	//fprintf(rpt, "\n<Ak,b>:\n");
	//WriteVector(rpt, 6, ab);
	MultiplyMatrixVector(6,6, invG, ab, p);
	//fprintf(rpt, "\nCoefficients:\n");
	//WriteVector(rpt, 6, p);
	//fprintf(rpt, "\nApproximating bock:\n");
	MakeApproxBlock(6,p, N, B);
	//for(k=0; k<N*N; k++) b[k]=B[k];
	//WriteMatrix(rpt, N,N, b);
	fprintf(rpt, "\n");
	CLOSEREPORT();
	break;
    }
  return Distance(A,B, N);
}

TAppRes Kernel::ApproximateImage(const RTImage img1, RTImage img2)
{
    ShowBegin();
    ShowMessage("Approximating Blocks...");
    ShowRatio(0);
    DWORD Time = GetTimeTicks();
    TAppRes Res;
    WORD wBlockCount = 0;

    img2.Copy(img1);
    FILE * rpt = fopen("params","w");
    double E, GlobalError = 0;
    double p[6];
    WORD N = B;
    BYTE * Block = new BYTE[N*N];
    BYTE * appBlock=new BYTE[N*N];
    WORD imax = img1.wHeight/N + (img1.wHeight%N ? 1:0);
    WORD jmax = img1.wWidth/N + (img1.wWidth%N ? 1:0);
    for (WORD ib=0; ib<imax; ib++)
      {
      ShowRatio(ib/(imax-1.0));
      for (WORD jb=0; jb<jmax; jb++)
	{
	  img1.GetBlock(ib*N, jb*N, N, Block);
	  E = Approximate(Block, appBlock, N, p);
	  if (E < ErrorThreshold)
	    {
	    wBlockCount++;
	    GlobalError += E;
	    // statistics on p...
	    WriteVector(rpt, 6, p); fprintf(rpt, "\n");
	    img2.SetBlock(ib*N, jb*N, N, appBlock);
	    }
	  // else img2.SetBlock(ib*N, jb*N, N, Block);
	}
      }
    delete Block;
    delete appBlock;
    fclose(rpt);

    Timing = GetTimeTicks()-Time;
    ShowEnd();

  Res.dblError = GlobalError;
  Res.dblThreshold = ErrorThreshold;
  Res.wCount = wBlockCount;
  return Res;
}

inline int Kernel::RangeCheck(double x) { return ( (int(x)>255) ? 255 : ( (x<0) ? 0 : int(x) ) ); }

#pragma argsused
#pragma off (unreferenced)
void Kernel::MakeApproxBlock(WORD m, LPDBL c, WORD N, PTBLOCK B)
{
  WORD I;
  for (WORD i=0; I=i*N,i<N; i++)
    for (WORD j=0; j<N; j++)
      B[I+j] = RangeCheck(c[0]+c[1]*i+c[2]*j+c[3]*i*i+c[4]*i*j+c[5]*j*j);
}
#pragma on (unreferenced)

BOOL ReadVector(FILE * in, WORD N, LPDBL v)
// We assume ROW vectors in the format { a1 , a2 , ... aN } -> v[0] ... v[N-1]
{
  fscanf(in," { ");
  for(int i=0; i<N-1; i++) if (fscanf(in,"%lf , ", &v[i])!=1) return FALSE;
  if (fscanf(in,"%lf } ", &v[N-1]) != 1) return FALSE;
  return TRUE;
}

BOOL ReadMatrix(FILE * in, WORD M, WORD N, LPDBL m)
// We assume matrix m to be stored and represented by ROW:
// { { a11, a12, ..., a1N },
//   { a21, a22, ..., a2N },     ->  m[0] m[1] ... m[N-1] m[N] .... ... m[(M-1)*N] ... m[M*N-1]
//   ....                            a11  a12       a1N   a21             aM1             aMN
//   { aM1, aM2, ..., aMN } }
{
  fscanf(in," { ");
  for(int i=0; i<M-1; i++)
    {
    if (!ReadVector(in, N, &m[i*N]))
      {
       char x1,x2,x3,x4,x5,x6,x7,x8;
       fscanf(in, "%c%c%c%c%c%c%c%c", &x1, &x2, &x3, &x4, &x5, &x6, &x7, &x8);
       MsgBox("File format error", "Offending char: %c%c%c%c%c%c%c%c\n%x %x %x %x %x %x %x %x",x1,x2,x3,x4,x5,x6,x7,x8,x1,x2,x3,x4,x5,x6,x7,x8);
       return FALSE;
      }
    fscanf(in," , ");
    }
  if (!ReadVector(in, N, &m[(M-1)*N])) return FALSE;
  fscanf(in," } ");
  return TRUE;
}

void WriteVector(FILE * out, WORD N, LPDBL v)
// We assume ROW vectors in the format { a1 , a2 , ... aN } <- v[0] ... v[N-1]
{
  fprintf(out,"{ ");
  for(int i=0; i<N-1; i++) fprintf(out,"%lf, ", v[i]);
  fprintf(out,"%lf }", v[N-1]);
}

void WriteMatrix(FILE * out, WORD M, WORD N, LPDBL m)
// We assume matrix m to be stored and represented by ROW:
// { { a11, a12, ..., a1N },
//   { a21, a22, ..., a2N },     <-  m[0] m[1] ... m[N-1] m[N] .... ... m[(M-1)*N] ... m[M*N-1]
//   ....                            a11  a12       a1N   a21             aM1             aMN
//   { aM1, aM2, ..., aMN } }
{
  fprintf(out,"{ ");
  for(int i=0; i<M-1; i++)
    {
    WriteVector(out, N, &m[i*N]);
    fprintf(out,",\n");
    }
  WriteVector(out, N, &m[(M-1)*N]);
  fprintf(out," }");
}

void NormalizeWeightVector(WORD N, LPDBL w)
// scale Weight Vector w, so that Sum(w[i],N) == N.
{
  int i;
  double sum = 0;
  for(i=0; i<N; i++) sum += w[i];
  for(i=0; i<N; i++) w[i] = w[i] / sum * N;
}

void MultiplyMatrixVector(WORD M, WORD N, LPDBL m, LPDBL v1, LPDBL v2)
// m is MxN (row x col) and v1, v2 are Nx1 even if they are represented by 1xN in memory
{
  for(int i=0,k; i<M; i++)
    for(v2[i]=0,k=0; k<N; k++) v2[i]+=m[i*N+k]*v1[k];
}

double DotVector(WORD N, LPDBL x, LPDBL y, LPDBL w)
// x,y are Nx1 vectors just as w we compute : <x,y> = Sum[w  x  y  ]
//                                                     k   k  k  k
{
  double ps = 0;
  if (w) for(int k=0; k<N; k++) ps+=w[k]*x[k]*y[k];
  else for(int k=0; k<N; k++) ps+=x[k]*y[k];
  return ps;
}

#pragma warn -pia
BOOL Kernel::ReadMathStream(char * filename, WORD deg, WORD coeff, WORD N /*, double * & aBasis, double * & aWeight, double * & ainvG */ )
// N*N is block size
{
  FILE * in;
  double hdr[3];
  if (Basis) delete Basis;
  if (Weight) delete Weight;
  if (invG) delete invG;
  Basis = new double[coeff*N*N];
  Weight= new double[N*N];
  invG  = new double[coeff*coeff];

  if (in=fopen(filename, "r"))
    {
/*
      BOOL b;
      b = ReadVector(in,3,hdr);
      MsgBox("","Status ok: %c\n Header: %lf, %lf, %lf ",(b?'Y':'N'), hdr[0], hdr[1], hdr[2]);
      b = ReadMatrix(in, coeff, N*N, Basis);
      MsgBox("","Status ok: %c\n Basis: %lf, %lf, %lf ",(b?'Y':'N'), Basis[0], Basis[1], Basis[2]);
      b = ReadVector(in, N*N, Weight);
      MsgBox("","Status ok: %c\n Weight: %lf, %lf, %lf ",(b?'Y':'N'), Weight[0], Weight[1], Weight[2]);
      b = ReadMatrix(in, coeff, coeff, invG);
      MsgBox("","Status ok: %c\n invG  : %lf, %lf, %lf ",(b?'Y':'N'),   invG[0],   invG[1],   invG[2]);
*/
      if ( !ReadVector(in, 3, hdr) ||
	   (hdr[0]!=deg || hdr[1]!=coeff || hdr[2]!=N) ||
	   !ReadMatrix(in, coeff, N*N, Basis) ||
	   !ReadVector(in, N*N, Weight) ||
	   !ReadMatrix(in, coeff, coeff, invG))
	return FALSE;
      else
	{
	  NormalizeWeightVector(N*N, Weight);
	  return TRUE;
	}

    }
  else return FALSE;
}
#pragma warn .pia

void Kernel::WriteHeader(FILE * cod, DWORD ID, WORD W, WORD H, WORD B, WORD did, WORD djd,
			     WORD idbits, WORD jdbits, WORD ENC_QUANTIZEALPHA,
			     WORD alphabits, WORD betabits, WORD polybits, WORD ENC_QUANTIZEPOLY)
{
    Pack(0, 0, cod);
    Pack(32, ID, cod);
    Pack(16, W, cod);
    Pack(16, H, cod);
    Pack(16, B, cod);
    Pack(16, did, cod);
    Pack(16, djd, cod);
    Pack(16, idbits, cod);
    Pack(16, jdbits, cod);
    Pack( 1, ENC_QUANTIZEALPHA, cod);
    if (ENC_QUANTIZEALPHA)
      {
      Pack(16, alphabits, cod);
      Pack(16, betabits, cod);
      }
    Pack(1, ENC_QUANTIZEPOLY, cod);
    Pack(16, polybits, cod); // poly bits
}

void Kernel::ReadHeader(FILE * cod, DWORD &ID, WORD &W, WORD &H, WORD &B, WORD &did, WORD &djd,
			     WORD &idbits, WORD &jdbits, WORD &DEC_QUANTIZEALPHA,
			     WORD &alphabits, WORD &betabits, WORD &polybits, WORD &DEC_QUANTIZEPOLY)
{
    ID = (DWORD) Unpack(32, cod);
    W  = (WORD)  Unpack(16, cod);
    H  = (WORD)  Unpack(16, cod);
    B  = (WORD)  Unpack(16, cod);
    did= (WORD)  Unpack(16, cod);
    djd= (WORD)  Unpack(16, cod);
    idbits = (WORD) Unpack(16, cod);
    jdbits = (WORD) Unpack(16, cod);
    DEC_QUANTIZEALPHA = (WORD) Unpack( 1, cod);
    if (DEC_QUANTIZEALPHA)
      {
      alphabits = (WORD) Unpack(16, cod);
      betabits  = (WORD) Unpack(16, cod);
      }
    else
      {
      alphabits = 0;
      betabits  = 0;
      }
    DEC_QUANTIZEPOLY = Unpack(1, cod); // ENC_QUANTIZEPOLY
    polybits = Unpack(16,cod); // poly bits

}


/*
|1| 3 |[[3]| fp1, fp2, ...,fp(nf)]|[c0[|c1[|c2[|c3..]..]]]|
 p  n    nf  floats                coefficients

bit sz
[1]1   0: polynomial coefficients
       1: IFS encoding
[2]3   0: coeff0=coeff1=...= 0
       1: coeff0!=0, coeff1=coeff2=...=0
       ...
       k: coeff0!=0,...,coeff(k-1)!=0, coeffk=coeff(k+1)=...=0
       ...
       7: floats follows with number of float in 3 bits
[
[4]3   number of floats that follow: nf
[6]32  1st float
       ....
       nf-th float
]
[4]polybits[0] : coefficient 0
....
[+]polybits[k] : coefficient k
*/


void Kernel::WritePolyCoeff(FILE * cod, double * p, int n)
{
   // Export these two defs in .INI file.
   double r[] =     { 127.0,  5,  5,  5,  5,  5 };  // half range: x :: -x..x
   DWORD qp[16], qpz[16];
   int polybits[] = {     8,  5,  5,  5,  5,  5 };
   int i,j;

   if (ENC_QUANTIZEPOLY)
     {
     for (i=0; i<n; i++)
       {
	 if (p[i]<=-r[i]) p[i]=-r[i];
	 if (p[i]>=r[i])  p[i]=r[i];
	 qp[i] = DWORD(0.5+(p[i]+r[i])/(2.0*r[i])*((1ul<<polybits[i])-1));
	 qpz[i]= DWORD(0.5+r[i]/(2.0*r[i])*((1ul<<polybits[i])-1));  // this is zero
	 p[i]  = double(qp[i])/double((1ul<<polybits[i])-1)*(2.0*r[i])-r[i];
       }
     // Recalculate Error, qError
     // if Error is still under threshold, puts coeffs into file

     if (FALSE) { Pack(1, 0, cod); Pack(3, 7, cod); for (i=0; i<n; i++) Pack((float)p[i],cod); }
     // or we may return with a bad status so that the encoding of this block will be done by IFS

     // Calculate exact number of coefficients not zero: -1,0,..,n-1
     j=n-1;  while (j>=0 && qp[j]==qpz[j]) j--;
     j++; // 0: coeff0==0, 1: only 1 coeff!=0, ...
     Pack(1, 0, cod);  // poly encoding
     Pack(3, j, cod);  // number of coeff. to write
     for(i=0; i<j; i++) Pack(polybits[i], qp[i], cod);
     }
   else
     { Pack(1,0,cod); Pack(3, 7, cod); for (i=0; i<n; i++) Pack((float)p[i],cod); }
} // WritePolyCoeff

void Kernel::ReadPolyCoeff(FILE * cod, double * p, int n, BOOL DEC_QUANTIZEPOLY)
{
   // Export these two defs in .INI file.
   double r[] =     { 127.0,  5,  5,  5,  5,  5 };  // half range: x :: -x..x
   DWORD qp[16], qpz[16];
   int polybits[] = {     8,  5,  5,  5,  5,  5 };
   int i,j;

   if (DEC_QUANTIZEPOLY)
     {
       j = Unpack(3, cod);  // number of coefficients
       if (j==7) for (i=0; i<6; i++) p[i]=Unpackfloat(cod);
       else
       for (i=0; i<j; i++)
       {
	 qp[i] = Unpack(polybits[i], cod);
	 p[i]  = double(qp[i])/double((1ul<<polybits[i])-1)*(2.0*r[i])-r[i];
       }
       for (i=j; i<n; i++) p[i]=0;
     }
   else
     { for(i=0; i<n; i++) p[i]=Unpackfloat(cod); }
} // ReadPolyCoeff


void Kernel::WriteQuantizedPoly(FILE * cod, double * p, int n)
{
  Pack(1, 0, cod);  // means: polynomial encoding follows
  int polyoverrange=FALSE, i;
  for (i=1; i<n; i++)
    if (p[i]<-5.0 || p[i]>5.0) polyoverrange=TRUE;
  if (ENC_QUANTIZEPOLY)
    {
    int i;
    if (polyoverrange)
      { // not quantized
      Pack(1, 0, cod);
      for(i=0; i<n; i++) Pack((float)p[i],cod);
      }
    else
      {
      int polint;
      double poldbl;
      Pack(1, 1, cod); // Valerio quantized
      poldbl = p[0] / 256.0; // Different quantization for p[0]
      if (poldbl < 0.0) poldbl = 0.0;
      if (poldbl > 1.0) poldbl = 1.0;
      polint = 0.5 + (poldbl + 1.0)/(2.0)*(1<<polybits);
      if (polint < 0) polint = 0;
      if (polint >= 1<<polybits) polint = (1<<polybits)-1;
      Pack(polybits, polint, cod);
      for (i=1; i<n; i++)
	{ // Valerio: non tiene conto del vero errore RMS
	poldbl = p[i];
	poldbl = (poldbl+5.0) / 10.0;
	if (poldbl < 0.0) poldbl = 0.0;
	if (poldbl > 1.0) poldbl = 1.0;
	polint = 0.5 + (poldbl + 1.0)/(2.0)*(1<<polybits);
	if (polint < 0) polint = 0;
	if (polint >= 1<<polybits) polint = (1<<polybits)-1;
	Pack(polybits, polint, cod);
	}
      }
    }
  else
    {  // Valerio
    for(i=0; i<n; i++) Pack((float)p[i],cod);
    }
}


void Kernel::MinimizeRms( TEncStats &S, PTBLOCK RBlock, PTBLOCK DBlock, WORD N, double SumR, double SumR2, double SumD, double SumD2,
				 double &alpha, double &beta, DWORD &ialpha, DWORD &ibeta, double &E0, double&E )
{
  double SumRD,Sum,det;
  SumRD = BlockCrossSum(RBlock, DBlock, N);
  Sum   = N*N,
  det   = Sum*SumD2-SumD*SumD;

  if (det == 0)
    {
    alpha = 0;
    beta = SumR/Sum;
    S.dwDet0++;
    }
  else
    {
    alpha = (Sum*SumRD - SumR*SumD)/det;
    beta  = (-SumD*SumRD + SumR*SumD2)/det;
    }

  if (alpha>1.0) S.dwAlphagt1++;
  if (alpha<0)   S.dwAlphalt0++;
  if (alpha<-1.0)S.dwAlphaltm1++;

  if (ENC_BOUNDALPHA)
  {
    if (alpha>alphamax) { alpha=alphamax; }
    if (alpha<alphamin) { alpha=alphamin; }
    beta = (SumR-alpha*SumD)/Sum;
  }

  E0=(alpha*(alpha*SumD2+2*beta*SumD-2*SumRD)+beta*(beta*Sum-2*SumR)+SumR2)/(N*N);
  if (ENC_QUANTIZEALPHA)
  {
    ialpha = DWORD(0.5 + (alpha-alphamin)/(alphamax-alphamin)*((1ul<<alphabits)-1));
    alpha  = double(ialpha)/double((1ul<<alphabits)-1)*(alphamax-alphamin)+alphamin;
    beta = (SumR-alpha*SumD)/Sum;
    if (beta>betamax) beta=betamax;
    if (beta<betamin) beta=betamin;
    ibeta  = DWORD(0.5 + (beta-betamin)/(betamax-betamin)*((1ul<<betabits)-1));
    beta   = double(ibeta)/((1ul<<betabits)-1)*(betamax-betamin)+betamin;
  }
  E=(alpha*(alpha*SumD2+2*beta*SumD-2*SumRD)+beta*(beta*Sum-2*SumR)+SumR2)/(N*N);
  if (E<0)
    { fprintf(rpt,"Error<0: %lf ", E);
      E = BlockError(DBlock, RBlock, N, alpha, beta)/(N*N);
      fprintf(rpt,"(%lf) [E0:%lf]\n",E,E0); fflush(rpt);
    }
}

void Kernel::ExtractChild(PTBLOCK Block, WORD N, PTBLOCK Child[4])
/*
 (0,0)->1 2
	3 4
*/
{
  WORD i2m,i2,I,M=N/2;
  for(register WORD i=0; i2=(i>>1)*M,i2m=((i/M)<<1),I=i*N,i<N; i++)
    for(register WORD j=0; j<N; j++)
      Child[i2m+j/M][(i%M)*M+j%M]=Block[I+j];
}

#pragma off (unreferenced)
void Kernel::ChildBlockEncode(WORD ir, WORD jr, WORD N, PTBLOCK Block)
{}
#pragma on (unreferenced)

double Kernel::BlockCrossSum(PTBLOCK A, PTBLOCK B, WORD N)
{
  double Sum = 0;
  register WORD i,j,I;
  for(i=0; I=i*N,i<N; i++)
    for(j=0; j<N; j++)
      Sum += Weight[I+j]*double(A[I+j])*double(B[I+j]);
  return Sum;
}

double Kernel::BlockError(PTBLOCK A, PTBLOCK B, WORD N,double alpha, double beta)
{
  double Sum = 0;
  register WORD i,j,I;
  for(i=0; I=i*N,i<N; i++)
    for(j=0; j<N; j++)
      Sum += Weight[I+j]*sqr(alpha*double(A[I+j])+beta-double(B[I+j]));
  return Sum;
}

#define bound(a,b) ( (int(a)<(b)) ? ((int(a)>0)?(int(a)):0) : (b) )

void Kernel::BlockLumScale(PTBLOCK A, PTBLOCK B, WORD N, double alpha, double beta)
{
  register WORD i,j,I;
  for(i=0; I=i*N,i<N; i++)
    for(j=0; j<N; j++)
      B[I+j] = bound(alpha*A[I+j]+beta,255);
}

void Kernel::BlockTransform(PTBLOCK A, PTBLOCK B, WORD N, TIsometries isometry)
{
  register WORD i,j,I,I1;
  switch (isometry)
  {
  case iso_IDENTITY:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+j];
     break;
  case iso_FLIPV:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+N-1-j];
     break;
  case iso_FLIPH:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+j];
     break;
  case iso_1stDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+i];
     break;
  case iso_2ndDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+(N-1-i)];
     break;
  case iso_ROT90:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+(N-1-i)];
     break;
  case iso_ROT180:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+(N-1-j)];
     break;
  case iso_ROT270:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+i];
     break;
  }
}

double Kernel::MSDifference(double * A, double * B, WORD N)
{
  double sum=0;
  WORD I;
  for(WORD i=0; I=i*N, i<N; i++)
    for(WORD j=0; j<N; j++)
      sum += sqr(A[I+j]-B[I+j]);
  return sum;
}

double Kernel::MSDifference(const RTImage img, WORD i0, WORD j0, BYTE * Block, WORD N)
{
  double sum=0;
  WORD I;
  for(WORD i=0; I=i*N, i<N; i++)
    for(WORD j=0; j<N; j++)
      sum += sqr(double(img.Pixel(i0+i,j0+j))-double(Block[I+j]));
  return sum;
}

double Kernel::sqrSum(const RTImage img)
{
  double sum=0;
  for(WORD i=0; i<img.wHeight; i++)
    for(WORD j=0; j<img.wWidth; j++) sum += sqr(double(img.Pixel(i,j)));
  return sum;
}


inline double dblRangeCheck(double x) { return ( (x)>255.0 ? 255.0 : ( x<0.0 ? 0.0 : (x) ) ); }

#pragma off (unreferenced)
void Kernel::MakeApproxBlock(WORD m, LPDBL c, WORD N, LPDBL B)
{
  WORD I;
  for (WORD i=0; I=i*N,i<N; i++)
    for (WORD j=0; j<N; j++)
      B[I+j] = dblRangeCheck(c[0]+c[1]*i+c[2]*j+c[3]*i*i+c[4]*i*j+c[5]*j*j);
}
#pragma on (unreferenced)

void Kernel::ConvertBlock(double    * A, BYTE    * B, WORD N)
{
  DWORD I;
  for (WORD i=0; I=DWORD(i)*N,i<N; i++)
    for (WORD j=0; j<N; j++)
      B[I+j] = RangeCheck(A[I+j]);
}

#pragma off (unreferenced)
void SetBlockToImage(double * Image, WORD W, WORD H, WORD i0, WORD j0, double    * Block, WORD N)
{
  DWORD I,I1;
  for (WORD i=0; I=DWORD(i)*N,I1=DWORD(i0+i)*W,i<N; i++)
    for (WORD j=0; j<N; j++)
      Image[I1+j0+j]=Block[I+j];
}

void GetBlockFromImage(double    * Image, WORD W, WORD H, WORD i0, WORD j0, double    * Block, WORD N)
{
  DWORD I,I1;
  for (WORD i=0; I=DWORD(i)*N,I1=DWORD(i0+i)*W,i<N; i++)
    for (WORD j=0; j<N; j++)
      Block[I+j]=Image[I1+j0+j];
}
#pragma on (unreferenced)

void Kernel::SetDibToImage(const RTImage img, double    * Image, WORD W, WORD H)
{
  DWORD I;
  for (WORD i=0; I=DWORD(i)*W,i<H; i++)
    for (WORD j=0; j<W; j++)
      Image[I+j]=img.Pixel(i,j);
}

void Kernel::Contract(double    * A, double    * B, WORD X)
// A is 2Nx2N -> B (N*N)
{
    DWORD N=X/2;
    DWORD M=X;
    for (WORD i=0; i<N; i++)
      for (WORD j=0; j<N; j++)
	{
	  WORD I = 2*i, J = 2*j;
	  B[i*N+j] = (A[I*M+J]+A[(I+1)*M+J]+A[I*M+J+1]+A[(I+1)*M+J+1])/4.0;
	}
}

void Kernel::BlockTransform(double    * A, double    * B, WORD N, TIsometries isometry)
{
  register DWORD i,j,I,I1;
  switch (isometry)
  {
  case iso_IDENTITY:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+j];
     break;
  case iso_FLIPV:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+N-1-j];
     break;
  case iso_FLIPH:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+j];
     break;
  case iso_1stDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+i];
     break;
  case iso_2ndDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+(N-1-i)];
     break;
  case iso_ROT90:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+(N-1-i)];
     break;
  case iso_ROT180:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+(N-1-j)];
     break;
  case iso_ROT270:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+i];
     break;
  }
}

// #define bound(a,b) ( ((a)<(b)) ? (((a)>0)?((a)):0) : (b) )
// #define bound(a,b) (a)

void Kernel::BlockLumScale(double    * A, double    * B, WORD N, double alpha, double beta)
{
  register DWORD i,j,I;
  for(i=0; I=DWORD(i)*N,i<N; i++)
    for(j=0; j<N; j++)
      B[I+j] = bound(alpha*A[I+j]+beta,255);
}


//#include "decode3.cpp"

void Kernel::MakeApproxBlock1(WORD m, LPDBL c, WORD M, WORD N, LPDBL B)
{
  WORD I;
  for (WORD i=0; I=i*N,i<N; i++)
    for (WORD j=0; j<N; j++)
      { double y = i*M*1.0/N;
	double x = j*M*1.0/N;
	B[I+j] = dblRangeCheck(c[0]+c[1]*y+c[2]*x+c[3]*y*y+c[4]*y*x+c[5]*x*x);
      }
}





/*
void Kernel::CalcMathStream()
{
  int i,j,k;
  int N = B;
  double ** basis;
  CalcBasis(0x03f, basis, N, N);
  typedef double * ptdouble;
  double ** A, ** B;
  A = new ptdouble[6];
  B = new ptdouble[6];
  for(i=0; i<6; i++) A[i] = new double[6], B[i] = new double[6];
  for(i=0; i<6; i++)
    for(j=0; j<6; j++)
      A[i][j] = Dot(N*N, basis[i], basis[j]);
  Inverse(A, B, N*N);
  // now comparing results with Mathematica calculated results.
  // Write out difference matrix.
  FILE * o = fopen("Math.rpt", "w");
  fprintf(o, "inv G:\n");
  for(i=0; i<6; i++) WriteVector(o, 6, B[i]), fprintf(o, "\n");
  fprintf(o, "Difference:\n");
  for(i=0; i<6; i++) for(j=0; j<6; j++) B[i][j] -= invG[6*i+j];
  for(i=0; i<6; i++) WriteVector(o, 6, B[i]), fprintf(o, "\n");
  fclose(o);


  for(i=0; i<6; i++) delete A[i], delete B[i], delete basis;
  delete A; delete B; delete basis;
}

*/

void BlockTransform(BYTE * A, BYTE * B, WORD N, TIsometries isometry)
{
  register WORD i,j,I,I1;
  switch (isometry)
  {
  case iso_IDENTITY:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+j];
     break;
  case iso_FLIPV:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I+N-1-j];
     break;
  case iso_FLIPH:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+j];
     break;
  case iso_1stDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+i];
     break;
  case iso_2ndDIAG:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+(N-1-i)];
     break;
  case iso_ROT90:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[j*N+(N-1-i)];
     break;
  case iso_ROT180:
     for(i=0; I=i*N,I1=(N-1-i)*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[I1+(N-1-j)];
     break;
  case iso_ROT270:
     for(i=0; I=i*N,i<N; i++)
       for(j=0; j<N; j++)
	 B[I+j] = A[(N-1-j)*N+i];
     break;
  }
}


void print(BYTE * a, int n, FILE * f)
{
  for(int k=0; k<n; k++) fprintf(f, "%d ",a[k]);
  fprintf(f, " [%2d] ",Permutation(a, n));
  // fprintf(f, "\n");
}

void findiso(BYTE * D, BYTE * R, int n, TIsometries iso[8], TIsometries iso1[8], int & m, int & m1)
{
  BYTE D1[10];
  int k,d1,d2,r1,r2;
  m=0; m1=0;
  for(k=0; k<n && R[k]!=0; k++);  r1=k;
  for(k=0; k<n && R[k]!=1; k++);  r2=k;
  TIsometries i;
  FORALLIsometries(i)
  {
    BlockTransform(D, D1, int(sqrt(n)), i);
    for(k=0; k<n && D1[k]!=0; k++); d1=k;
    for(k=0; k<n && D1[k]!=1; k++); d2=k;
    if (d1==r1 && d2==r2) iso[m++]=i;
    if (d1==r1) iso1[m1++]=i;
  }
}

void dojob(BYTE * R, int c, int n, FILE *f, BYTE IsometryClass[24][24] )
{ // we have Range with some pre-defined orientation
  // now we want find transformation wich maps D onto R with minimal error
  BYTE D[4], D1[4];
  int k;
  for (k=0; k<n; k++) D[k]=k;

  TIsometries iso[8], iso1[8];
  int m,m1,h=0;
  findiso(D,R,n, iso, iso1, m, m1);
  if (m)
    { fprintf(f,"R: "); print(R,n,f); fprintf(f," <-D: "); print(D,n,f); fprintf(f,"   ");
      for(k=0; k<m; k++) fprintf(f,"%d ",iso[k]);
      fprintf(f,"\n");
      IsometryClass[c][h++]=iso[0];
    }
  else
    { fprintf(f,"R: "); print(R,n,f); fprintf(f," <-D: "); print(D,n,f); fprintf(f," * ");
      for(k=0; k<m1; k++) fprintf(f,"%d ",iso1[k]);
      fprintf(f,"\n");
      IsometryClass[c][h++]=iso1[0];
    }
  while (Permute(D,n))
    {
       findiso(D,R,n, iso, iso1, m, m1);
       if (m)
	 { fprintf(f,"R: "); print(R,n,f); fprintf(f," <-D: "); print(D,n,f); fprintf(f,"   ");
	   for(k=0; k<m; k++) fprintf(f,"%d ",iso[k]);
	   for(k=0; k<m; k++)
	     {
	       BlockTransform(D,D1,2,iso[k]);
	       fprintf(f," {"); print(D1,n,f); fprintf(f,"},");
	     }
	   fprintf(f,"\n");
	   IsometryClass[c][h++]=iso[0];
	 }
       else
	 { fprintf(f,"R: "); print(R,n,f); fprintf(f," <-D: "); print(D,n,f); fprintf(f," * ");
	   for(k=0; k<m1; k++) fprintf(f,"%d ",iso1[k]);
	   for(k=0; k<m; k++)
	     {
	       BlockTransform(D,D1,2,iso1[k]);
	       fprintf(f," {"); print(D1,n,f); fprintf(f,"},");
	     }
	   fprintf(f,"\n");
	   IsometryClass[c][h++]=iso1[0];
	 }
    }
}

void Kernel::CalcScheme()
{
  ShowBegin();
  ShowMessage("Computing isometries...");
  ShowRatio(0);
  FILE * f = fopen("PERM","w");
  BYTE R[4];
  int i,j,k,h,N=4,c;
  for(k=0; k<N; k++) R[k]=k;
  for (i=0; i<24; i++)
    for (j=0; j<24; j++) IsometryClass[i][j]=iso_IDENTITY;

  h=1; k=1;
  for(j=1; j<=N; j++) h*=j;  // h=N!
  ShowRatio(1.0/h);
  c=0;
  dojob(R, c++, N, f, IsometryClass);
  while (Permute(R, N)) 
    {  
       dojob(R, c++, N, f, IsometryClass);
       ShowRatio((k++)*1.0/h);
    }

  fprintf(f, "\n    ");
  for (i=0; i<24; i++) fprintf(f, "%d ", i/10);
  fprintf(f, "\n    ");
  for (i=0; i<24; i++) fprintf(f, "%d ", i%10);
  fprintf(f, "\n");
  for (i=0; i<24; i++)
    {
    fprintf(f, "%2d: ",i);
    for (j=0; j<24; j++) fprintf(f, "%d ",IsometryClass[i][j]);
    fprintf(f,"\n");
    }

  ShowRatio(1.0);
  fclose(f);
  ShowEnd();
}


#pragma off (unreferenced)
void Kernel::InitEncStats(TEncStats & S, char * szInFN, char * szOutFN, WORD W, WORD H,
					 WORD nr, WORD mr, WORD nd, WORD md, WORD idbits, WORD jdbits,
					 DWORD ID, DWORD RPoolInfoSize, DWORD DPoolInfoSize, DWORD DPoolSize)
{
  memset(&S, 0x0, sizeof(TEncStats));
  S.dwIFSBlk = 0;
  S.dwPolyBlk= 0;
  S.dwSubBlk = 0;
  S.dwCountBlk=0;
  S.dwAlphagt1=0;   S.dwAppAlphagt1=0;
  S.dwDet0=0;       S.dwAppDet0=0;
  S.dwAlphalt0=0;   S.dwAppAlphalt0=0;
  S.dwAlphaltm1=0;  S.dwAppAlphaltm1=0;
  S.dwAlphaCount=0; S.dwAppAlpha=0;
  S.dwSearch=0;
  S.dwLSSsuccess=0; S.dwLSSsameiso=0;
  TIsometries iso;
  FORALLIsometries(iso) S.isofrequency[iso]=0;
  for (int i=0; i<15; i++) S.wErrRange[i]=0;
  for (i=0; i<100; i++) S.wAlpha[i]=0;
  for (i=0; i<100; i++) S.wBeta[i]=0;
  for (i=0; i<100; i++) S.wDeltax[i]=0;
  for (i=0; i<100; i++) S.wDeltay[i]=0;
  for (i=0; i<100; i++) S.wDist[i]=0;

  strcpy(S.szInFileName, szInFN);
  strcpy(S.szOutFileName, szOutFN);
  S.ENC_USEPOLY = ENC_USEPOLY;
  S.ENC_POLYONLY= ENC_POLYONLY;
  S.ENC_QUANTIZEALPHA = ENC_QUANTIZEALPHA;
  S.ENC_QUANTIZEPOLY  = ENC_QUANTIZEPOLY;
  S.ENC_BOUNDALPHA    = ENC_BOUNDALPHA;
  S.ENC_USEISOMETRIES = ENC_USEISOMETRIES;
  S.ENC_STRIPSHADES   = ENC_STRIPSHADES;
  S.ENC_OVERSTEP      = ENC_OVERSTEP;
  S.ENC_OVERSTEPUSEISOMETRIES = ENC_OVERSTEPUSEISOMETRIES;
  S.ENC_USEORIENTATIONSCHEME = ENC_USEORIENTATIONSCHEME;
  S.overstep = overstep;
  S.alphamin = alphamin;
  S.alphamax = alphamax;
  S.betamin = betamin;
  S.betamax = betamax;
  S.alphabits = alphabits;
  S.betabits = betabits;
  S.polybits = polybits;
  S.Metric = Metric;
  S.ClassScheme = ClassScheme;
  S.Approx = ApproximationType;
  S.EdgeThreshold = EdgeThreshold;
  S.MidrangeThreshold = MidrangeThreshold;
  S.ShadeThreshold = ShadeThreshold;
  if (Weight)
    S.w[0] = Weight[2*B], S.w[1] = Weight[2*B+1], S.w[2] = Weight[2*B+2];
  // true initialization
  S.B=B;
  S.W=W;
  S.H=H;
  S.did=did;
  S.djd=djd;
  S.nd=nd; S.md=md; S.nr=nr; S.mr=mr;
  //S.jdbits = jdbits;
  //S.idbits = idbits;
  //S.ID = ID;
  S.ErrorThreshold = ErrorThreshold;
  S.RIsize=RPoolInfoSize;
  S.DIsize=DPoolInfoSize;
  S.Dsize =DPoolSize;

  S.EncodeAlgorithm = EncodeAlgorithm;
  S.SimAnnSize = SimAnnSize;

  // Recursion support chunk
  for (i=0; i<3; i++) S.wSubBlock[4]=0;
  S.MaxLevel = MaxLevel;
  S.diameter = diameter;
  S.ErrorThresholdR = ErrorThresholdR;

}
#pragma on (unreferenced)

void Kernel::UpdateEncStats( TEncStats &S, TIsometries isoBest,
			     double alphaBest, double betaBest, double E,
			     WORD ir, WORD jr, WORD id, WORD jd)
{
  S.dwAppAlpha++;
  S.isofrequency[isoBest]++;
  if (alphaBest>1) S.dwAppAlphagt1++;
  if (alphaBest<0) S.dwAppAlphalt0++;
  if (alphaBest<-1)S.dwAppAlphaltm1++;
  int e = sqrt(E)*100/255.0;
  if (e<=2) S.wErrRange[0]++;
  else if (e<=5) S.wErrRange[1]++;
  else if (e<=10) S.wErrRange[2]++;
  else if (e<=15) S.wErrRange[3]++;
  else if (e<=20) S.wErrRange[4]++;
  else if (e<=100)S.wErrRange[(e-20)/10+5]++;
  else S.wErrRange[14]++;

  int i = (alphaBest-alphamin)/(alphamax-alphamin)*100; if (i>99) i=99; if (i<0) i=0;
  S.wAlpha[i]++;
  i = ((betaBest+256)/512.0)*100; if (i>99) i=99; if (i<0) i=0;
  S.wBeta[i]++;
  double dy = id*did,
	 dx = jd*djd,
	 ry = ir*B,
	 rx = jr*B;

  i = abs(dy-ry)/S.H*100.0; if (i>99) i=99; if (i<0) i=0;
  S.wDeltay[i]++;
  i = abs(dx-rx)/S.W*100.0; if (i>99) i=99; if (i<0) i=0;
  S.wDeltax[i]++;
  i = sqrt(sqr(dx-rx)+sqr(dy-ry))/sqrt(sqr(S.W*1.0)+sqr(S.H*1.0))*100;
  if (i>99) i=99; if (i<0) i=0;
  S.wDist[i]++;
}

void Kernel::FixupEncStats(TEncStats & S, double GlobalError, double GlobalErrorwoLSS)
{
  if (S.dwCountBlk) S.GlobalError = GlobalError/S.dwCountBlk;
  if (S.dwCountBlk) S.GlobalErrorwoLSS = GlobalErrorwoLSS/S.dwCountBlk;
  S.ElapsedTime = Timing;
}

void Kernel::WriteReportHeader( FILE * rpt, TEncStats & S, WORD idbits, WORD jdbits, DWORD ID)
{
  fprintf(rpt, "%s\n", GetTimeString());
  fprintf(rpt, "%s\nFilename: %s %dx%d\n", S.szOutFileName, S.szInFileName, S.W, S.H);
  fprintf(rpt, "Encoding Algorithm: %d\n", S.EncodeAlgorithm);
  fprintf(rpt, "Range: %dx%d (%dx%d:%lu)\nDomain: %dx%d step:(%d,%d) (%dx%d:%lu)\n",
	       S.B,S.B,S.nr,S.mr,DWORD(S.nr)*S.mr, D,D, S.did, S.djd, S.nd,S.md,DWORD(S.nd*S.md));
  fprintf(rpt, "Domain overstep: %d ", overstep);
  if (overstep)
    fprintf(rpt, " step(%d,%d) (%dx%d+%dx%d : %lu)\n",
		 S.did*S.overstep, S.djd*S.overstep, S.nd/S.overstep, S.md/S.overstep,
		 2*(S.overstep-1),2*(S.overstep-1),
		 (DWORD)(DWORD(S.nd/S.overstep*S.md/S.overstep)+sqr(2*(S.overstep-1)))  );
  fprintf(rpt, "Range pool size (info): %luK\nDomain pool size (info): %luK\nDomain pool size (bitmap): %luK\n",
	       S.RIsize/1024, S.DIsize/1024, S.Dsize/1024);
  fprintf(rpt, "Max recursion level: %d\n", S.MaxLevel);
  fprintf(rpt, "Diameter for sub-block search: %d\n", S.diameter);
  fprintf(rpt, "Error Threshold for partitions: %lf\n", S.ErrorThresholdR);
  fprintf(rpt, "Metric: %s\n", KrnlMetricsNames[S.Metric]);
  fprintf(rpt, "Approx: %s\n", KrnlApproxNames[S.Approx]);
  fprintf(rpt, "ErrorT: %lf\n", S.ErrorThreshold);
  fprintf(rpt, "CS    : %s\n", ClassSchemeNames[S.ClassScheme]);
  fprintf(rpt, "Edge T: %lf\nMidR T: %lf\nShadeT: %lf\n", S.EdgeThreshold, S.MidrangeThreshold, S.ShadeThreshold);
  fprintf(rpt, "Alphamin: %lf\nAlphamax: %lf\n", S.alphamin, S.alphamax);
  #define strBOOL(x) ((x)?("TRUE"):("FALSE"))
  fprintf(rpt, "Use polynomials: %s\nPolynomial only: %s\nBounded alpha: %s\nUse isometries: %s\n",
		 strBOOL(S.ENC_USEPOLY), strBOOL(S.ENC_POLYONLY), strBOOL(S.ENC_BOUNDALPHA), strBOOL(S.ENC_USEISOMETRIES));
  fprintf(rpt, "Use Local SubSearch: %s\nUse Isometries in LSS: %s\n",
		 strBOOL(S.ENC_OVERSTEP), strBOOL(S.ENC_OVERSTEPUSEISOMETRIES));
  fprintf(rpt, "Use orientation scheme: %s\n", strBOOL(S.ENC_USEORIENTATIONSCHEME));
  fprintf(rpt, "Simulated Annealing Size: %d\n", S.SimAnnSize);
  fprintf(rpt, "\nEncode ID: %0lX\n", ID);  // **** not in S ****
  fprintf(rpt, "idbits: %u\njdbits: %u\n", idbits, jdbits); // ***** not in S ****
  fprintf(rpt, "Quantize alpha, beta: %s\n", strBOOL(S.ENC_QUANTIZEALPHA));
  if (S.ENC_QUANTIZEALPHA)
    fprintf(rpt, "\tAlphaBits: %u\n\tBetaBits: %u\n\tPolyBits: %u\n", S.alphabits, S.betabits, S.polybits);
  fflush(rpt);
}

void Kernel::WriteReportFooter(TEncStats & S, double GlobalError, double GlobalErrorwoLSS)
{
  fprintf(rpt, "End of encoding.\nElapsed Time: %lu s  %d:%d:%d\n\n",
	       (DWORD)Timing/1000, int(Timing/3600000ul), int(((Timing/1000)%3600ul)/60),
	       int((((Timing/1000)%3600ul)%60)));
  fprintf(rpt, " Blocks     Poly     IFS   Searched   \n" \
	       "--------------------------------------\n" \
	     //" xxxxxx   xxxxxx  xxxxxx   xxxxxxxxxx
	     //" [xxx%]    [xx%]   [xx%]
	       " %6lu   %6lu  %6lu   %8lu\n",
	       S.dwCountBlk, S.dwPolyBlk, S.dwIFSBlk, S.dwSearch);
  if (S.dwCountBlk)
  fprintf(rpt, " [%3d%%]    [%2d%%]   [%2d%%]\n",
	       (int)100, int(S.dwPolyBlk*100/S.dwCountBlk),
	       int(S.dwIFSBlk*100/S.dwCountBlk));

  fprintf(rpt, "\n\nIFS encoding statistics:\n" \
	       " Search space  Alpha<-1   Alpha<0    Alpha>1    |Alpha|<1 \n" \
	       "----------------------------------------------------------\n" \
	     //" xxxxxxxx      xxxxxxxx   xxxxxxxx   xxxxxxxx   xxxxxxxx
	     //"   [xxx%]         [xx%]      [xx%]      [xx%]      [xx%]
	       " %8lu      %8lu   %8lu   %8lu   %8lu\n",
	       S.dwSearch, S.dwAlphaltm1, S.dwAlphalt0, S.dwAlphagt1,
	       S.dwSearch-S.dwAlphaltm1-S.dwAlphagt1);
  if (S.dwSearch)
  fprintf(rpt, "   [%3d%%]         [%2d%%]      [%2d%%]      [%2d%%]      [%2d%%]\n",
	       (int)100, int(S.dwAlphaltm1*100/S.dwSearch),
		int(S.dwAlphalt0*100/S.dwSearch),  int(S.dwAlphagt1*100/S.dwSearch),
		int((S.dwSearch-S.dwAlphaltm1-S.dwAlphagt1)*100/S.dwSearch) );

  fprintf(rpt, "\nDeterminant == 0 : %lu ", S.dwDet0 );
  if (S.dwSearch) fprintf(rpt, "[%d%%]\n", int(S.dwDet0*100/S.dwSearch));
  else fprintf(rpt,"\n");

  fprintf(rpt, "\nApplied blocks:\n" \
	       " Applied    Alpha<-1   Alpha<0    Alpha>1    |Alpha|<1 \n" \
	       "-------------------------------------------------------\n" \
	     //" xxxxxxxx   xxxxxxxx   xxxxxxxx   xxxxxxxx   xxxxxxxx
	     //"    [xx%]      [xx%]      [xx%]      [xx%]      [xx%]
	       " %8lu   %8lu   %8lu   %8lu   %8lu \n",
	       S.dwAppAlpha, S.dwAppAlphaltm1, S.dwAppAlphalt0, S.dwAppAlphagt1,
	       S.dwAppAlpha-S.dwAppAlphaltm1-S.dwAppAlphagt1 );
  if (S.dwAppAlpha)
  fprintf(rpt, "    [%2d%%]      [%2d%%]      [%2d%%]      [%2d%%]      [%2d%%]\n",
	       (int)100, int(S.dwAppAlphaltm1*100/S.dwAppAlpha),
		int(S.dwAppAlphalt0*100/S.dwAppAlpha), int(S.dwAppAlphagt1*100/S.dwAppAlpha),
		 int((S.dwAppAlpha-S.dwAppAlphaltm1-S.dwAppAlphagt1)*100/S.dwAppAlpha));

  fprintf(rpt, "\nIsometries:\n");
  fprintf(rpt, "IDENTITY FLIPV   FLIPH   1stDIAG 2ndDIAG ROT90   ROT180  ROT270\n" \
	       "---------------------------------------------------------------\n" );
	    // " xxxxxx  xxxxxx  xxxxxx  xxxxxx  xxxxxx  xxxxxx  xxxxxx  xxxxxx
	    // "  [xx%]   [xx%]   [xx%]   [xx%]   [xx%]   [xx%]   [xx%]   [xx%]
  TIsometries i;
  FORALLIsometries(i)
    fprintf(rpt," %6lu ",S.isofrequency[i]);
  fprintf(rpt,"\n");
  if (S.dwAppAlpha)
  FORALLIsometries(i)
    fprintf(rpt,"  [%2d%%] ",int(S.isofrequency[i]*100/S.dwAppAlpha));

  if (ENC_OVERSTEP && S.dwCountBlk)
  {
  fprintf(rpt, "\nLocal SubSearch success: %lu [%d%%]\n",
	       S.dwLSSsuccess, int(S.dwLSSsuccess*100/S.dwCountBlk));
  fprintf(rpt, "LSS change isometry: %lu [%d%%]\n",
	       S.dwCountBlk-S.dwLSSsameiso, int((S.dwCountBlk-S.dwLSSsameiso)*100/S.dwCountBlk));
  }

  fprintf(rpt, "\nSub Block encoding:\n");
  fprintf(rpt, "   1     1     2     4   \n"
	       "   8x8   4x4   4x4   4x4 \n"
	       "-------------------------\n");
	    // "  xxxx  xxxx  xxxx  xxxx \n"
  int k;
  for(k=0; k<4; k++) fprintf(rpt, "  %4u",S.wSubBlock[k]); fprintf(rpt,"\n");
  if (S.dwCountBlk)
    for(k=0; k<4; k++) fprintf(rpt, " [%2d%%]", int(S.wSubBlock[k]*100.0/S.dwCountBlk));
  fprintf(rpt, "\n");

  fprintf(rpt, "\n\nGlobal approximation error (sum of rms): %lf \n",GlobalError);
  if (S.dwCountBlk)
  fprintf(rpt, "       global rms error: %lf %d%% \n",
	       sqrt(GlobalError/S.dwCountBlk),int(sqrt(GlobalError/S.dwCountBlk)*100.0/255.0));
  if (S.dwCountBlk)
  fprintf(rpt, "       rms without LSS : %lf %d%% \n",
	       sqrt(GlobalErrorwoLSS/S.dwCountBlk),int(sqrt(GlobalErrorwoLSS/S.dwCountBlk)*100.0/255.0));

  fprintf(rpt, "Size of packed IFS file: %lu\n", S.packed);
  fprintf(rpt, "Size of original file  : %lu\n", S.original);
  fprintf(rpt, "Compression: %0.2f:1   bits/pixel:%0.4f\n", S.original*1.0/S.packed,
	       S.packed*8.0/S.original);

}
